/*
Atari 8-bit Rainbow
By Andrei Ellman
ellman@xs4all.nl
http://www.xs4all.nl/~ellman/ae-a

This program tries to use a mathematical formula to obtain the colours of the Atari 8-bit

This was compiled using the DJGPP C compiler and the Allegro package.

The DJGPP WWW homepage is:
http://www.delorie.com/djgpp/

The Allegro WWW homepage is:
http://www.talula.demon.co.uk/allegro/
*/


/*
This program tries to emulate the 256 colours of the Atari 8-bit. Unfortunately, I am using the wrong colourspace. In HLS space, all the colours come out too clean. I should be using YIQ or YUV space instead.

People (especially authours of emulators) should comment. What are reccomended gamma adjustments?
Have gamma adjustments as global variables?

Make the functions handle incorrect values (especially if the value must be from 0-1), but Hue can be anything. Perhaps values >1 can be rounded to 1 (mabey the luminamce adjustment may produce luminance>1 if lumshift is large enough).

In the declarations for the values from 0-1, should I make them 'unsigned'?
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <allegro.h>

#define MAX_INTENSITY 63.0
#define UNDEFINED -999.0


PALLETE pallete;	/* Gets changed during interrupts */

void atari2rgb(RGB *, int, int);
void hls2rgb(float, float, float, RGB *);
void colourcycle_until_keypressed();
void colourcycle();



/* hls_to_rgb:
 *  Converts from HLS colorspace to RGB values.
 *  Taken From Foley & VanDam Fig. 13.37 with a few modifications.
 *  Input: h[0.0,6.0 or UNDEFINED] s[0.0,1.0] v[0.0,1.0]
 *  (s must be 0.0 when h is UNDEFINED)
 *  Output: An RGB structure. each RGB in [0,MAX_INTENSITY]
 *  Note: This does not check paramaters for the valid ranges,
 */

/* First, a helper function for working out a value for a colour chanel */
float hls_to_rgb_value (float n1, float n2, float hue)
{
	if (hue > 6.0)
		hue -= 6.0;
	else if (hue < 0.0)
		hue += 6.0;

	if (hue < 1.0)
		return n1 + (n2-n1)*hue;
	else if (hue < 3.0)
		return n2;
	else if (hue < 4.0)
		return n1 + (n2-n1)*(4.0-hue);
	else
		return n1;
}

/* The function proper */
void hls_to_rgb (float h, float l, float s, RGB *rgb)
{
	float m1, m2;

	/*
	AE: The commented out line appeared in Foley&VanDam
	The line below is it my attempt to correct it. I think it works.
	*/
	/* m2 = (l<=0.5) ? (l*(l+s)) : (l+s - (l*s)); */
	m2 = (l<=0.5) ? (l+(l*s)) : (l+s - (l*s));
	m1 = 2.0 * l - m2;

	if (s == 0.0) {   /* greyscale (h should ideally be UNDEFINED) */
    	rgb->r = rgb->g = rgb->b = (unsigned char)(l * MAX_INTENSITY);
	} else {
		rgb->r = (unsigned char)(hls_to_rgb_value(m1, m2, h+2.0) * MAX_INTENSITY);
		rgb->g = (unsigned char)(hls_to_rgb_value(m1, m2, h) * MAX_INTENSITY);
		rgb->b = (unsigned char)(hls_to_rgb_value(m1, m2, h-2.0) * MAX_INTENSITY);
	}
}	/* End of hls_to_rgb */


void main()
{

	int h, l;
	int y, c;	// Loop counters

	allegro_init();
	install_keyboard();
	set_gfx_mode(GFX_AUTODETECT, 640, 480, 0, 0);

	set_pallete(black_pallete);

	/* Make the palette */

	for (h=0; h<=15; h++) {
		for (l=0; l<=15; l++) {
			atari2rgb(&pallete[h*16+l], h, l);
		}
	}

	set_pallete(pallete);

	/* Draw the lines */

	for (y=0; y<480; y++) {
		hline(screen, 0, y, 639, y % 256);
	}


	colourcycle_until_keypressed();

	for (c=255; c>0; c--)
	{
		circlefill(screen, SCREEN_W/2, SCREEN_H/2, c, c);
	}

	colourcycle_until_keypressed();

}

/* This function maps atari colours to RGB colours.
First it converts them to HLS colours, and then calls a function to convert HLS to RGB colours.
'xl_hue' and 'xl_luminance' are from 0-15
Remember that in most Atari graphics modes, only 8 levels of luminance were alowed (ie. the even numbers).
Perhaps have another function that adjusts the saturation levels, gamma, the hue shift
and the extrema of the intensity values for coloured colours.*/

void atari2rgb(RGB *the_rgbvalue, int xl_hue, int xl_luminance)

{
	const float hueshift = 5.4;
	/* Clockwise shift of the Hue to map the atari Hue values with the HLS hue values. */
	const float lummul = 0.7; // was 0.7;	// Range of luminance values
	const float lumshift = (1.0 - lummul) / 2.0;	// This can also be offset by a contant.
		/* lummul and lumshift are used to define extrema for the luminance values of colours.
		Otherwise, all hues with luminance of 0 would be black, or white with a luminance of 15. */
	const float huedegrees = 6.0/15.0;
		/* The number of HLS units of hue per one Atari hue value. This should allways be the same. 6 is circumference of hue circle, 15 is the number of non-black hues in the Atari spectrum.*/

	if (xl_hue==0) {
		/* For hue=0 (black) */
		hls_to_rgb(0.0, (float)xl_luminance / 15.0, 0.0, the_rgbvalue);
	} else {
		/* For coloured hue values (1-15) */
		hls_to_rgb(6.0 - fmod ( (((float)(xl_hue-1))*huedegrees + hueshift), 6.0), ((float)xl_luminance / 15.0)*lummul + lumshift, 1.0, the_rgbvalue);
		// hls_to_rgb(((float)(xl_hue-1))*huedegrees, (float)xl_luminance / 15.0, 1.0, the_rgbvalue);
	}

}

void colourcycle_until_keypressed()
{
	/* animate the image by rotating the pallete */

	clear_keybuf();

	while (!keypressed())

	{
		colourcycle();
	}

}

void colourcycle()
{
	static RGB cycle_temp;
	static int cycle_i, cycle_j;

	cycle_temp = pallete[255];
	for (cycle_i=255, cycle_j=254; cycle_i>0;) {
		pallete[cycle_i--] = pallete[cycle_j--];
	}
	pallete[0] = cycle_temp;
	set_pallete(pallete);	// Implies vsync()
}
